@@ -13,8 +13,15 @@ from absl.testing import parameterized
 from MuyGPyS.examples.classify import do_classify
 from MuyGPyS.examples.two_class_classify_uq import do_classify_uq, do_uq
 from MuyGPyS.examples.regress import do_regress
+from MuyGPyS.examples.fast_regress import do_fast_regress
 from MuyGPyS.gp.muygps import MuyGPS, MultivariateMuyGPS as MMuyGPS
-from MuyGPyS.optimize.objective import mse_fn
+from MuyGPyS.optimize.loss import mse_fn
+
+from MuyGPyS._src.mpi_utils import (
+    _consistent_chunk_tensor,
+    _consistent_unchunk_tensor,
+    _consistent_reduce_scalar,
+)
 
 
 class ClassifyAPITest(parameterized.TestCase):
@@ -26,6 +33,7 @@ class ClassifyAPITest(parameterized.TestCase):
         nn_count: int,
         batch_count: int,
         loss_method: str,
+        obj_method: str,
         opt_method: str,
         nn_kwargs: Dict,
         k_kwargs: Union[Dict, Union[List[Dict], Tuple[Dict, ...]]],
@@ -47,6 +55,7 @@ class ClassifyAPITest(parameterized.TestCase):
             nn_count,
             batch_count,
             loss_method,
+            obj_method,
             opt_method,
             nn_kwargs,
             kern,
@@ -100,6 +109,7 @@ class ClassifyAPITest(parameterized.TestCase):
         nn_count: int,
         batch_count: int,
         loss_method: str,
+        obj_method: str,
         opt_method: str,
         nn_kwargs: Dict,
         kern: Optional[str],
@@ -122,6 +132,7 @@ class ClassifyAPITest(parameterized.TestCase):
             nn_count=nn_count,
             batch_count=batch_count,
             loss_method=loss_method,
+            obj_method=obj_method,
             opt_method=opt_method,
             kern=kern,
             k_kwargs=k_kwargs,
@@ -156,9 +167,24 @@ class ClassifyAPITest(parameterized.TestCase):
             )
 
         predicted_labels = np.argmax(surrogate_predictions, axis=1)
-        acc = np.mean(predicted_labels == np.argmax(test["output"], axis=1))
+        target_labels = np.argmax(test["output"], axis=1)
+        test_count = len(target_labels)
         if np.all(np.unique(train["output"]) == np.unique([-1.0, 1.0])):
             predicted_labels = 2 * predicted_labels - 1
+            target_labels = 2 * target_labels - 1
+
+        target_labels = _consistent_chunk_tensor(target_labels)
+        correct_count = np.count_nonzero(predicted_labels == target_labels)
+
+        correct_count = _consistent_reduce_scalar(correct_count)
+        acc = correct_count / test_count
+
+        surrogate_predictions = _consistent_unchunk_tensor(
+            surrogate_predictions
+        )
+        predicted_labels = _consistent_unchunk_tensor(predicted_labels)
+        crosswise_dists = _consistent_unchunk_tensor(crosswise_dists)
+        pairwise_dists = _consistent_unchunk_tensor(pairwise_dists)
         return (
             classifier,
             surrogate_predictions,
@@ -177,6 +203,7 @@ class ClassifyAPITest(parameterized.TestCase):
         opt_batch_count: int,
         uq_batch_count: int,
         loss_method: str,
+        obj_method: str,
         opt_method: str,
         uq_objectives: Union[List[Callable], Tuple[Callable, ...]],
         nn_kwargs: Dict,
@@ -197,6 +224,7 @@ class ClassifyAPITest(parameterized.TestCase):
             opt_batch_count,
             uq_batch_count,
             loss_method,
+            obj_method,
             opt_method,
             uq_objectives,
             nn_kwargs,
@@ -251,6 +279,7 @@ class ClassifyAPITest(parameterized.TestCase):
         opt_batch_count: int,
         uq_batch_count: int,
         loss_method: str,
+        obj_method: str,
         opt_method: str,
         uq_objectives: Union[List[Callable], Tuple[Callable, ...]],
         nn_kwargs: Dict,
@@ -266,6 +295,7 @@ class ClassifyAPITest(parameterized.TestCase):
             opt_batch_count=opt_batch_count,
             uq_batch_count=uq_batch_count,
             loss_method=loss_method,
+            obj_method=obj_method,
             opt_method=opt_method,
             uq_objectives=uq_objectives,
             k_kwargs=k_kwargs,
@@ -290,6 +320,7 @@ class RegressionAPITest(parameterized.TestCase):
         nn_count: int,
         batch_count: int,
         loss_method: str,
+        obj_method: str,
         opt_method: str,
         sigma_method: Optional[str],
         variance_mode: Optional[str],
@@ -314,6 +345,7 @@ class RegressionAPITest(parameterized.TestCase):
             nn_count,
             batch_count,
             loss_method,
+            obj_method,
             opt_method,
             sigma_method,
             variance_mode,
@@ -375,6 +407,7 @@ class RegressionAPITest(parameterized.TestCase):
         nn_count: int,
         batch_count: int,
         loss_method: str,
+        obj_method: str,
         opt_method: str,
         sigma_method: Optional[str],
         variance_mode: Optional[str],
@@ -393,6 +426,7 @@ class RegressionAPITest(parameterized.TestCase):
         np.ndarray,
         np.ndarray,
     ]:
+        # print("gets here")
         ret = do_regress(
             test["input"],
             train["input"],
@@ -400,6 +434,7 @@ class RegressionAPITest(parameterized.TestCase):
             nn_count=nn_count,
             batch_count=batch_count,
             loss_method=loss_method,
+            obj_method=obj_method,
             opt_method=opt_method,
             sigma_method=sigma_method,
             variance_mode=variance_mode,
@@ -450,6 +485,12 @@ class RegressionAPITest(parameterized.TestCase):
                 )
         else:
             raise ValueError(f"Variance mode {variance_mode} is not supported.")
+
+        predictions = _consistent_unchunk_tensor(predictions)
+        variance = _consistent_unchunk_tensor(variance)
+        crosswise_dists = _consistent_unchunk_tensor(crosswise_dists)
+        pairwise_dists = _consistent_unchunk_tensor(pairwise_dists)
+
         mse = mse_fn(predictions, test["output"])
         return (
             regressor,
@@ -459,3 +500,109 @@ class RegressionAPITest(parameterized.TestCase):
             crosswise_dists,
             pairwise_dists,
         )
+
+
+class FastRegressionAPITest(parameterized.TestCase):
+    def _do_fast_regress_test_chassis(
+        self,
+        train: Dict[str, np.ndarray],
+        test: Dict[str, np.ndarray],
+        target_mse: float,
+        nn_count: int,
+        batch_count: int,
+        loss_method: str,
+        obj_method: str,
+        opt_method: str,
+        sigma_method: Optional[str],
+        variance_mode: Optional[str],
+        nn_kwargs: Dict,
+        k_kwargs: Union[Dict, Union[List[Dict], Tuple[Dict, ...]]],
+        opt_kwargs: Dict,
+        kern: Optional[str] = None,
+        apply_sigma_sq: bool = False,
+        return_distances: bool = False,
+        verbose: bool = False,
+    ) -> None:
+        (regressor, predictions, mse) = self._do_fast_regress(
+            train,
+            test,
+            nn_count,
+            batch_count,
+            loss_method,
+            obj_method,
+            opt_method,
+            sigma_method,
+            variance_mode,
+            nn_kwargs,
+            k_kwargs,
+            opt_kwargs,
+            kern=kern,
+            apply_sigma_sq=apply_sigma_sq,
+            return_distances=return_distances,
+            verbose=verbose,
+        )
+        self.assertEqual(predictions.shape, test["output"].shape)
+        print(f"obtains mse: {mse}")
+        self.assertLessEqual(mse, target_mse)
+
+        param_names, param_vals, _ = regressor.get_optim_params()
+        if len(param_names) > 0:
+            print("finds hyperparameters:")
+            for i, p in enumerate(param_names):
+                print(f"\t{p} : {param_vals[i]}")
+
+    def _do_fast_regress(
+        self,
+        train: Dict[str, np.ndarray],
+        test: Dict[str, np.ndarray],
+        nn_count: int,
+        batch_count: int,
+        loss_method: str,
+        obj_method: str,
+        opt_method: str,
+        sigma_method: Optional[str],
+        variance_mode: Optional[str],
+        nn_kwargs: Dict,
+        k_kwargs: Union[Dict, Union[List[Dict], Tuple[Dict, ...]]],
+        opt_kwargs: Dict,
+        kern: Optional[str] = None,
+        apply_sigma_sq: bool = True,
+        return_distances: bool = False,
+        verbose: bool = False,
+    ) -> Tuple[
+        Union[MuyGPS, MMuyGPS],
+        np.ndarray,
+        float,
+        np.ndarray,
+        np.ndarray,
+        np.ndarray,
+    ]:
+        (
+            regressor,
+            _,
+            predictions,
+            precomputed_coefficient_matrix,
+            _,
+        ) = do_fast_regress(
+            test["input"],
+            train["input"],
+            train["output"],
+            nn_count=nn_count,
+            batch_count=batch_count,
+            loss_method=loss_method,
+            obj_method=obj_method,
+            opt_method=opt_method,
+            sigma_method=sigma_method,
+            kern=kern,
+            k_kwargs=k_kwargs,
+            nn_kwargs=nn_kwargs,
+            opt_kwargs=opt_kwargs,
+            verbose=verbose,
+        )
+
+        mse = mse_fn(predictions, test["output"])
+        return (
+            regressor,
+            predictions,
+            mse,
+        )
