@@ -10,6 +10,7 @@ MuyGPs implementation
 import numpy as np
 
 from typing import Callable, Dict, List, Optional, Tuple, Union
+from MuyGPyS.gp.distance import crosswise_distances
 
 from MuyGPyS.gp.kernels import (
     _get_kernel,
@@ -20,30 +21,23 @@ from MuyGPyS.gp.kernels import (
 
 from MuyGPyS import config
 
-if config.muygpys_jax_enabled is False:  # type: ignore
-    from MuyGPyS._src.gp.numpy_distance import _make_regress_tensors
-    from MuyGPyS._src.gp.numpy_muygps import (
-        _muygps_compute_solve,
-        _muygps_compute_diagonal_variance,
-        _muygps_sigma_sq_optim,
-    )
-else:
-    from MuyGPyS._src.gp.jax_distance import _make_regress_tensors
+from MuyGPyS._src.gp.distance import (
+    _make_regress_tensors,
+    _make_fast_regress_tensors,
+)
 
-    # Presently, these implementations appear to be slower than numpy on CPU
-    # (but much faster on GPU)
-    if config.muygpys_gpu_enabled is False:  # type: ignore
-        from MuyGPyS._src.gp.numpy_muygps import (
-            _muygps_compute_solve,
-            _muygps_compute_diagonal_variance,
-            _muygps_sigma_sq_optim,
-        )
-    else:
-        from MuyGPyS._src.gp.jax_muygps import (
-            _muygps_compute_solve,
-            _muygps_compute_diagonal_variance,
-            _muygps_sigma_sq_optim,
-        )
+
+from MuyGPyS._src.gp.distance.numpy import (
+    _make_regress_tensors as _make_regress_tensors_n,
+)
+from MuyGPyS._src.gp.muygps import (
+    _muygps_compute_solve,
+    _muygps_compute_diagonal_variance,
+    _muygps_fast_regress_solve,
+    _muygps_fast_regress_precompute,
+)
+from MuyGPyS._src.mpi_utils import _is_mpi_mode
+from MuyGPyS.optimize.utils import _switch_on_opt_method
 
 
 class MuyGPS:
@@ -193,7 +187,7 @@ class MuyGPS:
                 the `(nn_count, nn_count` -shaped kernel matrices corresponding
                 to each of the batch elements.
             Kcross:
-                A tensor of shape `(batch_count, nn_count)` containing the
+                A matrix of shape `(batch_count, nn_count)` containing the
                 `1 x nn_count` -shaped cross-covariance matrix corresponding
                 to each of the batch elements.
             batch_nn_targets:
@@ -227,8 +221,8 @@ class MuyGPS:
                 the `(nn_count, nn_count` -shaped kernel matrices corresponding
                 to each of the batch elements.
             Kcross:
-                A tensor of shape `(batch_count, nn_count)` containing the
-                `1 x nn_count` -shaped cross-covariance matrix corresponding
+                A matrix of shape `(batch_count, nn_count)` containing the
+                `1 x nn_count` -shaped cross-covariance vector corresponding
                 to each of the batch elements.
             eps:
                 The value of the homoscedastic nugget parameter.
@@ -249,6 +243,7 @@ class MuyGPS:
         variance_mode: Optional[str] = None,
         apply_sigma_sq: bool = True,
         return_distances: bool = False,
+        indices_by_rank: bool = False,
     ) -> Union[
         np.ndarray,
         Tuple[np.ndarray, np.ndarray],
@@ -292,11 +287,14 @@ class MuyGPS:
                 nearest neighbor sets and a `(test_count, nn_count, nn_count)`
                 tensor containing the pairwise distances between the test data's
                 nearest neighbor sets.
+            indices_by_rank:
+                If `True`, construct the tensors using local indices with no
+                communication. Only for use in MPI mode.
 
         Returns
         -------
         responses:
-            A matrix of shape `(batch_count, response_count,)` whose rows are
+            A matrix of shape `(batch_count, response_count)` whose rows are
             the predicted response for each of the given indices.
         diagonal_variance:
             A vector of shape `(batch_count,)` consisting of the diagonal
@@ -308,16 +306,17 @@ class MuyGPS:
             distance of the corresponding test element to each of its nearest
             neighbors. Only returned if `return_distances is True`.
         pairwise_dists:
-            A tensor of shape `(test_count, nn_count, nn_count,)` whose latter
+            A tensor of shape `(test_count, nn_count, nn_count)` whose latter
             two dimensions contain square matrices containing the pairwise
             distances between the nearest neighbors of the test elements. Only
             returned if `return_distances is True`.
         """
-        (
-            crosswise_dists,
-            pairwise_dists,
-            batch_nn_targets,
-        ) = _make_regress_tensors(
+        tensor_fn = (
+            _make_regress_tensors_n
+            if _is_mpi_mode() is True and indices_by_rank is True
+            else _make_regress_tensors
+        )
+        (crosswise_dists, pairwise_dists, batch_nn_targets,) = tensor_fn(
             self.kernel.metric, indices, nn_indices, test, train, targets
         )
         K = self.kernel(pairwise_dists)
@@ -338,6 +337,67 @@ class MuyGPS:
                 responses, variances = responses
                 return responses, variances, crosswise_dists, pairwise_dists
 
+    def build_fast_regress_coeffs(
+        self,
+        train: np.ndarray,
+        nn_indices: np.ndarray,
+        targets: np.ndarray,
+        indices_by_rank: bool = False,
+    ) -> np.ndarray:
+        """
+        Produces coefficient matrix for fast regression given in Equation
+        (8) of [dunton2022fast]_. To form each row of this matrix, we compute
+
+        .. math::
+            \\mathbf{C}_{N^*}(i, :) =
+                (K_{\\hat{\\theta}} (X_{N^*}, X_{N^*})
+                + \\varepsilon I_k)^{-1} Y(X_{N^*}).
+
+        Here :math:`X_{N^*}` is the union of the nearest neighbor of the ith
+        test point and the `nn_count - 1` nearest neighbors of this nearest
+        neighbor, :math:`K_{\\hat{\\theta}}` is the trained kernel functor
+        specified by `self.kernel`, :math:`\\varepsilon I_k` is a diagonal
+        homoscedastic noise matrix whose diagonal is the value of the
+        `self.eps` hyperparameter, and :math:`Y(X_{N^*})` is the
+        `(train_count,)` vector of responses corresponding to the
+        training features indexed by $N^*$.
+
+        Args:
+            train:
+                The full training data matrix of shape
+                `(train_count, feature_count)`.
+            nn_indices:
+                The nearest neighbors indices of each
+                training points of shape `(train_count, nn_count)`.
+            targets:
+                A matrix of shape `(train_count, response_count)` whose rows are
+                vector-valued responses for each training element.
+        Returns:
+            A matrix of shape `(train_count, nn_count)` whose rows are
+            the precomputed coefficients for fast regression.
+
+        """
+        (
+            pairwise_dists_fast,
+            train_nn_targets_fast,
+        ) = _make_fast_regress_tensors(
+            self.kernel.metric, nn_indices, train, targets
+        )
+        K = self.kernel(pairwise_dists_fast)
+
+        return self._build_fast_regress_coeffs(
+            K, self.eps(), train_nn_targets_fast
+        )
+
+    @staticmethod
+    def _build_fast_regress_coeffs(
+        K: np.ndarray,
+        eps: float,
+        train_nn_targets_fast: np.ndarray,
+    ) -> np.ndarray:
+
+        return _muygps_fast_regress_precompute(K, eps, train_nn_targets_fast)
+
     def regress(
         self,
         K: np.array,
@@ -398,12 +458,12 @@ class MuyGPS:
                 the `(nn_count, nn_count` -shaped kernel matrices corresponding
                 to each of the batch elements.
             Kcross:
-                A tensor of shape `(batch_count, nn_count)` containing the
+                A matrix of shape `(batch_count, nn_count)` containing the
                 `1 x nn_count` -shaped cross-covariance matrix corresponding
                 to each of the batch elements.
             batch_nn_targets:
-                A tensor of shape `(batch_count, nn_count, response_count)` whose
-                last dimension lists the vector-valued responses for the
+                A tensor of shape `(batch_count, nn_count, response_count)`
+                whose last dimension lists the vector-valued responses for the
                 nearest neighbors of each batch element.
             variance_mode:
                 Specifies the type of variance to return. Currently supports
@@ -416,7 +476,7 @@ class MuyGPS:
         Returns
         -------
         responses:
-            A matrix of shape `(batch_count, response_count,)` whose rows are
+            A matrix of shape `(batch_count, response_count)` whose rows are
             the predicted response for each of the given indices.
         diagonal_variance:
             A vector of shape `(batch_count,)` consisting of the diagonal
@@ -464,9 +524,151 @@ class MuyGPS:
                 f"Variance mode {variance_mode} is not implemented."
             )
 
-    def get_opt_fn(self) -> Callable:
+    def fast_regress_from_indices(
+        self,
+        indices: np.ndarray,
+        nn_indices: np.ndarray,
+        test_features: np.ndarray,
+        train_features: np.ndarray,
+        closest_index: np.ndarray,
+        coeffs_mat: np.ndarray,
+    ) -> np.ndarray:
         """
-        Return a regress function for use in optimization.
+        Performs fast regression using provided
+        cross-covariance, the index of the training point closest to the
+        queried test point, and precomputed coefficient matrix.
+
+        Returns the predicted response in the form of a posterior
+        mean for each element of the batch of observations, as computed in
+        Equation (9) of [dunton2022fast]_. For each test point
+        :math:`\\mathbf{z}`, we compute
+
+        .. math::
+            \\widehat{Y} (\\mathbf{z} \\mid X) =
+                K_\\theta (\\mathbf{z}, X_{N^*}) \mathbf{C}_{N^*}.
+
+        Here :math:`X_{N^*}` is the union of the nearest neighbor of the queried
+        test point :math:`\\mathbf{z}` and the nearest neighbors of that
+        training point, :math:`K_\\theta` is the kernel functor specified
+        by `self.kernel`, and :math:`\mathbf{C}_{N^*}` is the matrix of
+        precomputed coefficients given in Equation (8) of [dunton2022fast]_.
+
+        Args:
+            indices:
+                A vector of shape `('batch_count,)` providing the indices of the
+                test features to be queried in the formation of the crosswise
+                distance tensor.
+            nn_indices:
+                A matrix of shape `('batch_count, nn_count)` providing the index
+                of the closest training point to each queried test point, as
+                well as the `nn_count - 1` closest neighbors of that point.
+            test_features:
+                A matrix of shape `(batch_count, feature_count)` containing
+                the test data points.
+            train_features:
+                A matrix of shape `(train_count, feature_count)` containing the
+                training data.
+            closest_index:
+                A vector of shape `('batch_count,)` for which each
+                entry is the index of the training point closest to
+                each queried point.
+            coeffs_mat:
+                A matrix of shape `('batch_count, nn_count)` providing
+                precomputed coefficients for fast regression.
+
+        Returns:
+            A matrix of shape `(batch_count,)` whose rows are
+            the predicted response for each of the given indices.
+        """
+        crosswise_dists = crosswise_distances(
+            test_features,
+            train_features,
+            indices,
+            nn_indices,
+        )
+
+        Kcross = self.kernel(crosswise_dists)
+        return self.fast_regress(
+            Kcross,
+            coeffs_mat[closest_index, :],
+        )
+
+    def fast_regress(
+        self,
+        Kcross: np.ndarray,
+        coeffs_mat: np.ndarray,
+    ) -> np.ndarray:
+        """
+        Performs fast regression using provided
+        cross-covariance and precomputed coefficient matrix.
+
+        Assumes that cross-covariance matrix `Kcross` is already computed and
+        given as an argument. To implicitly construct these values from indices
+        instead use :func:`~MuyGPyS.gp.muygps.MuyGPS.fast_regress_from_indices`.
+
+        Returns the predicted response in the form of a posterior
+        mean for each element of the batch of observations, as computed in
+        Equation (9) of [dunton2022fast]_. For each test point
+        :math:`\\mathbf{z}`, we compute
+
+        .. math::
+            \\widehat{Y} (\\mathbf{z} \\mid X) =
+                K_\\theta (\\mathbf{z}, X_{N^*}) \mathbf{C}_{N^*}.
+
+        Here :math:`X_{N^*}` is the union of the nearest neighbor of the queried
+        test point :math:`\\mathbf{z}` and the nearest neighbors of that
+        training point, :math:`K_\\theta` is the kernel functor
+        specified by `self.kernel`, and :math:`\mathbf{C}_{N^*}` is
+        the matrix of precomputed coefficients given in Equation (8)
+        of [dunton2022fast]_.
+
+        Args:
+            Kcross:
+                A matrix of shape `(batch_count, nn_count)` containing the
+                `1 x nn_count` -shaped cross-covariance vector corresponding
+                to each of the batch elements.
+            coeffs_mat:
+                A matrix of shape `(batch_count, nn_count)` whose rows
+                are given by precomputed coefficients for fast regression.
+
+
+        Returns:
+            A matrix of shape `(batch_count, response_count)` whose rows are
+            the predicted response for each of the given indices.
+        """
+        return self._fast_regress(
+            Kcross,
+            coeffs_mat,
+        )
+
+    @staticmethod
+    def _fast_regress(
+        Kcross: np.ndarray,
+        coeffs_mat: np.ndarray,
+    ) -> np.ndarray:
+        responses = _muygps_fast_regress_solve(Kcross, coeffs_mat)
+        return responses
+
+    def get_opt_mean_fn(self, opt_method) -> Callable:
+        """
+        Return a posterior mean function for use in optimization.
+
+        This function is designed for use with
+        :func:`MuyGPyS.optimize.chassis.optimize_from_tensors()`. The
+        `opt_method` parameter determines the format of the returned function.
+
+        Returns:
+            A function implementing a posterior mean, where `eps` is either
+            fixed or takes updating values during optimization. The format of
+            the function depends upon `opt_method`.
+        """
+        return _switch_on_opt_method(
+            opt_method, self.get_kwargs_opt_mean_fn, self.get_array_opt_mean_fn
+        )
+
+    def get_array_opt_mean_fn(self) -> Callable:
+        """
+        Return a posterior mean function for use in optimization.
 
         This function is designed for use with
         :func:`MuyGPyS.optimize.chassis.optimize_from_tensors()` with
@@ -475,16 +677,18 @@ class MuyGPS:
         element or is not included.
 
         Returns:
-            A function implementing regression, where `eps` is either fixed or
-            takes updating values during optimization. The function expects a
+            A function implementing posterior mean, where `eps` is either fixed
+            or takes updating values during optimization. The function expects a
             list of current hyperparameter values for unfixed parameters, which
             are expected to occur in a certain order matching how they are set
             in `~MuyGPyS.gp.muygps.MuyGPS.get_optim_params()`.
         """
-        return self._get_opt_fn(_muygps_compute_solve, self.eps)
+        return self._get_array_opt_mean_fn(_muygps_compute_solve, self.eps)
 
     @staticmethod
-    def _get_opt_fn(solve_fn: Callable, eps: Hyperparameter) -> Callable:
+    def _get_array_opt_mean_fn(
+        solve_fn: Callable, eps: Hyperparameter
+    ) -> Callable:
         if not eps.fixed():
 
             def caller_fn(K, Kcross, batch_nn_targets, x0):
@@ -497,9 +701,9 @@ class MuyGPS:
 
         return caller_fn
 
-    def get_kwargs_opt_fn(self) -> Callable:
+    def get_kwargs_opt_mean_fn(self) -> Callable:
         """
-        Return a regress function for use in optimization.
+        Return a posterior mean function for use in optimization.
 
         This function is designed for use with
         :func:`MuyGPyS.optimize.chassis.optimize_from_tensors()` with
@@ -507,15 +711,17 @@ class MuyGPS:
         via a keyword argument or not at all.
 
         Returns:
-            A function implementing regression, where `eps` is either fixed or
-            takes updating values during optimization. The function expects
-            keyword arguments corresponding to current hyperparameter values for
-            unfixed parameters.
+            A function implementing the posterior mean, where `eps` is either
+            fixed or takes updating values during optimization. The function
+            expects keyword arguments corresponding to current hyperparameter
+            values for unfixed parameters.
         """
-        return self._get_kwargs_opt_fn(_muygps_compute_solve, self.eps)
+        return self._get_kwargs_opt_mean_fn(_muygps_compute_solve, self.eps)
 
     @staticmethod
-    def _get_kwargs_opt_fn(solve_fn: Callable, eps: Hyperparameter) -> Callable:
+    def _get_kwargs_opt_mean_fn(
+        solve_fn: Callable, eps: Hyperparameter
+    ) -> Callable:
         if not eps.fixed():
 
             def caller_fn(K, Kcross, batch_nn_targets, **kwargs):
@@ -528,48 +734,64 @@ class MuyGPS:
 
         return caller_fn
 
-    def sigma_sq_optim(
-        self,
-        K: np.ndarray,
-        nn_indices: np.ndarray,
-        targets: np.ndarray,
-    ) -> np.ndarray:
+    def get_opt_var_fn(self, opt_method) -> Callable:
         """
-        Optimize the value of the :math:`\\sigma^2` scale parameter for each
-        response dimension.
+        Return a posterior variance function for use in optimization.
 
-        We approximate :math:`\\sigma^2` by way of averaging over the analytic
-        solution from each local kernel.
-
-        .. math::
-            \\sigma^2 = \\frac{1}{bk} * \\sum_{i \\in B}
-                        Y_{nn_i}^T K_{nn_i}^{-1} Y_{nn_i}
-
-        Here :math:`Y_{nn_i}` and :math:`K_{nn_i}` are the target and kernel
-        matrices with respect to the nearest neighbor set in scope, where
-        :math:`k` is the number of nearest neighbors and :math:`b = |B|` is the
-        number of batch elements considered.
-
-        Args:
-            K:
-                A tensor of shape `(batch_count, nn_count, nn_count)` containing
-                the `(nn_count, nn_count` -shaped kernel matrices corresponding
-                to each of the batch elements.
-            nn_indices:
-                An integral matrix of shape `(batch_count, nn_count)` listing the
-                nearest neighbor indices for all observations in the test batch.
-            targets:
-                A matrix of shape `(batch_count, response_count)` whose rows list
-                the vector-valued responses for all of the training targets.
+        This function is designed for use with
+        :func:`MuyGPyS.optimize.chassis.optimize_from_tensors()`. The
+        `opt_method` parameter determines the format of the returned function.
 
         Returns:
-            A vector of shape `(response_count)` listing the value of sigma^2
-            for each dimension.
+            A function implementing posterior variance, where `eps` is either
+            fixed or takes updating values during optimization. The format of
+            the function depends upon `opt_method`.
         """
-        self.sigma_sq._set(
-            _muygps_sigma_sq_optim(K, nn_indices, targets, self.eps())
+        return _switch_on_opt_method(
+            opt_method, self.get_kwargs_opt_var_fn, self.get_array_opt_var_fn
         )
-        return self.sigma_sq()
+
+    def get_array_opt_var_fn(self) -> Callable:
+        return self._get_array_opt_var_fn(
+            _muygps_compute_diagonal_variance, self.eps
+        )
+
+    @staticmethod
+    def _get_array_opt_var_fn(
+        var_fn: Callable, eps: Hyperparameter
+    ) -> Callable:
+        if not eps.fixed():
+
+            def caller_fn(K, Kcross, x0):
+                return var_fn(K, Kcross, x0[-1])
+
+        else:
+
+            def caller_fn(K, Kcross, x0):
+                return var_fn(K, Kcross, eps())
+
+        return caller_fn
+
+    def get_kwargs_opt_var_fn(self) -> Callable:
+        return self._get_kwargs_opt_var_fn(
+            _muygps_compute_diagonal_variance, self.eps
+        )
+
+    @staticmethod
+    def _get_kwargs_opt_var_fn(
+        var_fn: Callable, eps: Hyperparameter
+    ) -> Callable:
+        if not eps.fixed():
+
+            def caller_fn(K, Kcross, **kwargs):
+                return var_fn(K, Kcross, kwargs["eps"])
+
+        else:
+
+            def caller_fn(K, Kcross, **kwargs):
+                return var_fn(K, Kcross, eps())
+
+        return caller_fn
 
 
 class MultivariateMuyGPS:
@@ -641,71 +863,6 @@ class MultivariateMuyGPS:
         """
         return bool(np.all([model.fixed() for model in self.models]))
 
-    def sigma_sq_optim(
-        self,
-        pairwise_dists: np.ndarray,
-        nn_indices: np.ndarray,
-        targets: np.ndarray,
-    ) -> np.ndarray:
-        """
-        Optimize the value of the :math:`\\sigma^2` scale parameter for each
-        response dimension.
-
-        We approximate :math:`\\sigma^2` by way of averaging over the analytic
-        solution from each local kernel.
-
-        .. math::
-            \\sigma^2 = \\frac{1}{n} * Y^T  K^{-1}  Y
-
-        Args:
-            pairwise_dists:
-                A tensor of shape `(batch_count, nn_count, nn_count)` containing
-                the `(nn_count, nn_count)` -shaped pairwise nearest neighbor
-                distance matrices corresponding to each of the batch elements.
-            nn_indices:
-                An integral matrix of shape `(batch_count, nn_count)` listing the
-                nearest neighbor indices for all observations in the testing
-                batch.
-            targets:
-                A matrix of shape `(train_count, response_count)` whose rows
-                are the responses for each training element.
-
-        Returns:
-            A vector of shape `(response_count,)` listing the found value of
-            :math:`\\sigma^2` for each response dimension.
-        """
-
-        self.sigma_sq._set(
-            self._sigma_sq_optim(
-                self.models, pairwise_dists, nn_indices, targets
-            )
-        )
-        return self.sigma_sq()
-
-    @staticmethod
-    def _sigma_sq_optim(
-        models: List[MuyGPS],
-        pairwise_dists: np.ndarray,
-        nn_indices: np.ndarray,
-        targets: np.ndarray,
-    ) -> np.ndarray:
-        batch_count, nn_count = nn_indices.shape
-        target_count, response_count = targets.shape
-        if response_count != len(models):
-            raise ValueError(
-                f"Response count ({response_count}) does not match the number "
-                f"of models ({len(models)})."
-            )
-
-        K = np.zeros((batch_count, nn_count, nn_count))
-        sigma_sqs = np.zeros((response_count,))
-        for i, muygps in enumerate(models):
-            K = muygps.kernel(pairwise_dists)
-            sigma_sqs[i] = muygps.sigma_sq_optim(
-                K, nn_indices, targets[:, i].reshape(target_count, 1)
-            )
-        return sigma_sqs
-
     def regress_from_indices(
         self,
         indices: np.ndarray,
@@ -716,6 +873,7 @@ class MultivariateMuyGPS:
         variance_mode: Optional[str] = None,
         apply_sigma_sq: bool = True,
         return_distances: bool = False,
+        indices_by_rank: bool = False,
     ) -> Union[
         np.ndarray,
         Tuple[np.ndarray, np.ndarray],
@@ -758,10 +916,13 @@ class MultivariateMuyGPS:
                 nearest neighbor sets and a `(test_count, nn_count, nn_count)`
                 tensor containing the pairwise distances between the test data's
                 nearest neighbor sets.
+            indices_by_rank:
+                If `True`, construct the tensors using local indices with no
+                communication. Only for use in MPI mode.
         Returns
         -------
         responses:
-            A matrix of shape `(batch_count, response_count,)` whose rows are
+            A matrix of shape `(batch_count, response_count)` whose rows are
             the predicted response for each of the given indices.
         variance:
             A vector of shape `(batch_count,)` consisting of the diagonal
@@ -772,16 +933,18 @@ class MultivariateMuyGPS:
             distance of the corresponding test element to each of its nearest
             neighbors. Only returned if `return_distances is True`.
         pairwise_dists:
-            A tensor of shape `(test_count, nn_count, nn_count,)` whose latter
+            A tensor of shape `(test_count, nn_count, nn_count)` whose latter
             two dimensions contain square matrices containing the pairwise
             distances between the nearest neighbors of the test elements. Only
             returned if `return_distances is True`.
         """
-        (
-            crosswise_dists,
-            pairwise_dists,
-            batch_nn_targets,
-        ) = _make_regress_tensors(
+        tensor_fn = (
+            _make_regress_tensors_n
+            if _is_mpi_mode() is True and indices_by_rank is True
+            else _make_regress_tensors
+        )
+
+        (crosswise_dists, pairwise_dists, batch_nn_targets,) = tensor_fn(
             self.metric,
             indices,
             nn_indices,
@@ -888,7 +1051,7 @@ class MultivariateMuyGPS:
         Returns
         -------
         responses:
-            A matrix of shape `(batch_count, response_count,)` whose rows are
+            A matrix of shape `(batch_count, response_count)` whose rows are
             the predicted response for each of the given indices.
         diagonal_variance:
             A vector of shape `(batch_count, response_count)` consisting of the
@@ -943,3 +1106,193 @@ class MultivariateMuyGPS:
         if variance_mode == "diagonal":
             return responses, diagonal_variance
         return responses
+
+    def build_fast_regress_coeffs(
+        self,
+        train: np.ndarray,
+        nn_indices: np.ndarray,
+        targets: np.ndarray,
+        indices_by_rank: bool = False,
+    ) -> np.ndarray:
+        """
+        Produces coefficient tensor for fast regression given in Equation
+        (8) of [dunton2022fast]_. To form the tensor, we compute
+
+        .. math::
+            \\mathbf{C}_{N^*}(i, :, j) =
+                (K_{\\hat{\\theta_j}} (X_{N^*}, X_{N^*}) +
+                \\varepsilon I_k)^{-1} Y(X_{N^*}).
+
+        Here :math:`X_{N^*}` is the union of the nearest neighbor of the ith
+        test point and the `nn_count - 1` nearest neighbors of this nearest
+        neighbor, :math:`K_{\\hat{\\theta_j}}` is the trained kernel functor
+        corresponding the jth response and specified by `self.models`,
+        :math:`\\varepsilon I_k` is a diagonal homoscedastic noise matrix whose
+        diagonal  is the value of the `self.eps` hyperparameter,
+        and :math:`Y(X_{N^*})` is the `(train_count, response_count)`
+        matrix of responses corresponding to the training features indexed
+        by $N^*$.
+
+        Args:
+            train:
+                The full training data matrix of shape
+                `(train_count, feature_count)`.
+            nn_indices:
+                The nearest neighbors indices of each
+                training points of shape `(train_count, nn_count)`.
+            targets:
+                A matrix of shape `(train_count, response_count)` whose rows are
+                vector-valued responses for each training element.
+        Returns:
+            A tensor of shape `(batch_count, nn_count, response_count)`
+            whose entries comprise the precomputed coefficients for fast
+            regression.
+
+        """
+        (
+            pairwise_dists_fast,
+            train_nn_targets_fast,
+        ) = _make_fast_regress_tensors(self.metric, nn_indices, train, targets)
+
+        return self._build_fast_regress_coeffs(
+            self.models, pairwise_dists_fast, train_nn_targets_fast
+        )
+
+    @staticmethod
+    def _build_fast_regress_coeffs(
+        models: List[MuyGPS],
+        pairwise_dists_fast: np.ndarray,
+        train_nn_targets_fast: np.ndarray,
+    ) -> np.ndarray:
+        train_count, nn_count, response_count = train_nn_targets_fast.shape
+        coeffs_mat = np.zeros((train_count, nn_count, response_count))
+        for i, model in enumerate(models):
+            K = model.kernel(pairwise_dists_fast)
+            coeffs_mat[:, :, i] = _muygps_fast_regress_precompute(
+                K, model.eps(), train_nn_targets_fast[:, :, i]
+            )
+
+        return coeffs_mat
+
+    def fast_regress_from_indices(
+        self,
+        indices: np.ndarray,
+        nn_indices: np.ndarray,
+        test_features: np.ndarray,
+        train_features: np.ndarray,
+        closest_index: np.ndarray,
+        coeffs_mat: np.ndarray,
+    ) -> np.ndarray:
+        """
+        Performs fast multivariate regression using provided
+        vectors and matrices used in constructed the crosswise distances matrix,
+        the index of the training point closest to the queried test point,
+        and precomputed coefficient matrix.
+
+        Returns the predicted response in the form of a posterior
+        mean for each element of the batch of observations, as computed in
+        Equation (9) of [dunton2022fast]_. For each test point
+        :math:`\\mathbf{z}`, we compute
+
+        .. math::
+            \\widehat{Y} (\\mathbf{z} \\mid X) =
+                K_\\theta (\\mathbf{z}, X_{N^*}) \mathbf{C}_{N^*}.
+
+        Here :math:`X_{N^*}` is the union of the nearest neighbor of the queried
+        test point :math:`\\mathbf{z}` and the nearest neighbors of that
+        training point, :math:`K_\\theta` is the kernel functor specified
+        by `self.kernel`, and :math:`\mathbf{C}_{N^*}` is the matrix of
+        precomputed coefficients given in Equation (8) of [dunton2022fast]_.
+
+        Args:
+            indices:
+                A vector of shape `('batch_count,)` providing the indices of the
+                test features to be queried in the formation of the crosswise
+                distance tensor.
+            nn_indices:
+                A matrix of shape `('batch_count, nn_count)` providing the index
+                of the closest training point to each queried test point, as
+                well as the `nn_count - 1` closest neighbors of that point.
+            test_features:
+                A matrix of shape `(batch_count, feature_count)` containing
+                the test data points.
+            train_features:
+                A matrix of shape `(train_count, feature_count)` containing the
+                training data.
+            closest_index:
+                A vector of shape `(batch_count,)` for which each entry is
+                the index of the training point closest to each queried
+                test point.
+            coeffs_mat:
+                A tensor of shape `(batch_count, nn_count, response_count)`
+                providing the precomputed coefficients for fast regression.
+
+        Returns:
+            A matrix of shape `(batch_count, response_count)` whose rows are
+            the predicted response for each of the given indices.
+        """
+
+        crosswise_dists = crosswise_distances(
+            test_features,
+            train_features,
+            indices,
+            nn_indices,
+        )
+
+        return self.fast_regress(
+            crosswise_dists,
+            coeffs_mat[closest_index, :, :],
+        )
+
+    def fast_regress(
+        self,
+        crosswise_dists: np.ndarray,
+        coeffs_mat: np.ndarray,
+    ) -> np.ndarray:
+        """
+        Performs fast regression using provided
+        crosswise distances and precomputed coefficient matrix.
+
+        Returns the predicted response in the form of a posterior
+        mean for each element of the batch of observations, as computed in
+        Equation (9) of [dunton2022fast]_. For each test point
+        :math:`\\mathbf{z}`, we compute
+
+        .. math::
+            \\widehat{Y} (\\mathbf{z} \\mid X) =
+                K_\\theta (\\mathbf{z}, X_{N^*}) \mathbf{C}_{N^*}.
+
+        Here :math:`X_{N^*}` is the union of the nearest neighbor of the queried
+        test point :math:`\\mathbf{z}` and the nearest neighbors of that
+        training point, :math:`K_\\theta` is the kernel functor specified by
+        `self.kernel`, and :math:`\mathbf{C}_{N^*}` is the matrix of
+        precomputed coefficients given in Equation (8) of [dunton2022fast]_.
+
+        Args:
+            crosswise_dists:
+                A matrix of shape `(batch_count, nn_count)` whose rows list the
+                distance of the corresponding test element to each of its
+                nearest neighbors.
+            coeffs_mat:
+                A tensor of shape `(batch_count, nn_count, response_count)`
+                providing the precomputed coefficients for fast regression.
+
+
+        Returns:
+            A matrix of shape `(batch_count, response_count)` whose rows are
+            the predicted response for each of the given indices.
+        """
+        models = self.models
+        responses = self._fast_regress(models, crosswise_dists, coeffs_mat)
+        return responses
+
+    @staticmethod
+    def _fast_regress(
+        models: List[MuyGPS],
+        crosswise_dists: np.ndarray,
+        coeffs_mat: np.ndarray,
+    ) -> np.ndarray:
+        Kcross = np.zeros(coeffs_mat.shape)
+        for i, model in enumerate(models):
+            Kcross[:, :, i] = model.kernel(crosswise_dists)
+        return _muygps_fast_regress_solve(Kcross, coeffs_mat)
